from src.Mucklet.ResClient import ResClient
from src.Mucklet.ResClient import Request
from src.Mucklet.log import Logger
from src.Mucklet.types import Character
from src.Mucklet.types import Room
from src.Mucklet.types import Exit
from src.Mucklet.types import EventBase
from src.Mucklet.types import CharacterMessageEvent
from src.Mucklet.types import CharacterPoseableMessageEvent
from src.Mucklet.types import TargetedCharacterMessageEvent
from src.Mucklet.types import TargetRoomMessageEvent
from src.Mucklet.types import TargetedCharacterEvents
import time
import sys


class Bot:
	def __init__(self, token: str, client: ResClient):
		self.token = token
		self.client = client
		self.rid = None
		self.id = None
		self.name = None
		self.surname = None
		self.full_name = None
		self.description = None
		self.booted = False
		self.log_info = Logger((75, 75, 200))
		self.log_error = Logger((200, 75, 75), True)
		self.log_success = Logger((75, 200, 75))
		self.on_calls = {"say"           : [],
		                 "pose"          : [],
		                 "wakeup"        : [],
		                 "sleep"         : [],
		                 "leave"         : [],
		                 "arrive"        : [],
		                 "describe"      : [],
		                 "action"        : [],
		                 
		                 "ooc"           : [],
		                 
		                 "whisper"       : [],
		                 "message"       : [],
		                 "warn"          : [],
		                 "mail"          : [],
		                 "address"       : [],
		                 "controlRequest": [],
		                 
		                 "travel"        : [],
		                 
		                 "summon"        : [],
		                 "join"          : [],
		                 "leadRequest"   : [],
		                 "followRequest" : [],
		                 "follow"        : [],
		                 "stopFollow"    : [],
		                 "stopLead"      : [], }
		
		self.match_types = {"say"           : CharacterMessageEvent,
		                    "pose"          : CharacterMessageEvent,
		                    "wakeup"        : CharacterMessageEvent,
		                    "sleep"         : CharacterMessageEvent,
		                    "leave"         : CharacterMessageEvent,
		                    "arrive"        : CharacterMessageEvent,
		                    "describe"      : CharacterMessageEvent,
		                    "action"        : CharacterMessageEvent,
		                    
		                    "ooc"           : CharacterPoseableMessageEvent,
		                    
		                    "whisper"       : TargetedCharacterMessageEvent,
		                    "message"       : TargetedCharacterMessageEvent,
		                    "warn"          : TargetedCharacterMessageEvent,
		                    "mail"          : TargetedCharacterMessageEvent,
		                    "address"       : TargetedCharacterMessageEvent,
		                    "controlRequest": TargetedCharacterMessageEvent,
		                    
		                    "travel"        : TargetRoomMessageEvent,
		                    
		                    "summon"        : TargetedCharacterEvents,
		                    "join"          : TargetedCharacterEvents,
		                    "leadRequest"   : TargetedCharacterEvents,
		                    "followRequest" : TargetedCharacterEvents,
		                    "follow"        : TargetedCharacterEvents,
		                    "stopFollow"    : TargetedCharacterEvents,
		                    "stopLead"      : TargetedCharacterEvents, }
		
		self.waiting_for_response_from = {}  # dict[character: time_sent]
	
	def get_error(self, response: Request, kill: bool = True) -> bool:
		# if there is no error then return False
		if "error" not in response.received_data:
			return False
		# get the error
		error = response.received_data["error"]
		
		# set up the string to log
		string = f"Error: \n"
		
		# if there is a code then add it to the string
		if "code" in error:
			string += f"\t\tCode: {error['code']}\n"
		
		# if there is a message then add it to the string
		if "message" in error:
			string += f"\t\tMessage: {error['message']}\n"
		
		# log the string
		self.log_error(string)
		
		# if the kill flag is set to True then kill the bot
		if kill:
			self.kill()
		
		# if kill flag is set to False then return True
		return True
	
	def get_version(self) -> str:
		self.log_info("Retrieving protocol version")
		version = self.client.version("1.2.1").wait()
		self.get_error(version)
		self.log_success(f"Successfully retrieved protocol version - {version.value().get('result').get('protocol')}")
		return version.value().get('result').get('protocol')
	
	def authenticate(self) -> None:
		self.log_info(f"Authenticating bot - {self.token:.5}...")
		auth = self.client.auth("auth", "authenticateBot",
		                        token = self.token).wait()
		self.get_error(auth)
		self.log_success("Successfully authenticated bot")
	
	def get_bot(self) -> None:
		self.log_info(f"Retrieving bot - {self.token:.5}...")
		bot = self.client.call("core", "getBot").wait()
		self.get_error(bot)
		
		self.rid = bot.value().get("result").get("rid")
		data = bot.value().get("result").get("models").get(
				bot.value().get("result").get("models").get(self.rid).get("char").get("rid")
		)
		
		self.name = data.get("name")
		self.surname = data.get("surname")
		self.full_name = f"{self.name} {self.surname}"
		self.description = data.get("desc")
		self.id = data.get("id")
		self.log_success(f"Successfully retrieved bot - {self.full_name}")
	
	def control_bot(self) -> None:
		self.log_info("Controlling bot")
		control = self.client.call("core", "bot", self.id, "controlChar",
		                           charId = self.id).wait()
		self.get_error(control, False)
		self.log_success("Successfully controlled bot")
	
	def subscribe_to_all(self) -> None:
		self.client.subscribe("core", "info").send()
		self.client.subscribe("tag", "info").send()
		self.client.subscribe("mail", "info").send()
		self.client.subscribe("note", "info").send()
		self.client.subscribe("report", "info").send()
		self.client.subscribe("support", "info").send()
		self.client.subscribe("client", "web", "info").send()
		self.client.subscribe("core", "nodes").send()
		self.client.subscribe("tags", "tags").send()
		self.client.subscribe("tags", "groups").send()
		self.client.subscribe("core", "chars", "awake").send()
		
	def on(self, *event: str):
		def decorator(func):
			for e in event:
				if e not in self.on_calls:
					self.log_error(f"Unknown event type - {e}")
					continue
				if func in self.on_calls[e]:
					self.log_error(f"Function already registered - {func}")
					continue
				self.on_calls[e].append(func)
		
		return decorator
	
	def boot(self):
		@self.client.on("message")
		def on_message(message: dict):
			received_data = message
			
			if "id" in received_data:
				return
			
			if "data" not in received_data:
				self.log_error(f"Error - No Data - {received_data}")
				return
			
			if "event" not in received_data:
				self.log_error(f"Error - No Event - {received_data}")
				return
			
			if "type" not in received_data["data"]:
				return
			
			data = received_data["data"]
			
			if "targets" in data:
				self.log_info(f"Recieved multiple targets - {data.get('targets')}")
			
			# now we want to convert the dictionary into its own nested classes
			# create a new EventBase object
			event = EventBase(id = data.get("id"), type = data.get("type"),
			                  time = data.get("time"), sig = data.get("sig"))
			
			# now we want to work out what type of event it is
			if event.type not in self.on_calls:
				self.log_error(f"Unknown event type - {event.type}")
				return
			
			# now we want to create the event object
			event_class = self.match_types[event.type]
			
			if event_class == CharacterMessageEvent:
				event.event = CharacterMessageEvent(
						character = Character(id = data.get("char", {}).get("id"),
						                      bot = self),
						message = data.get("msg"),
						puppeteer = Character(id = data.get("puppeteer", {}).get("id"),
						                      bot = self))
			
			elif event_class == CharacterPoseableMessageEvent:
				event.event = CharacterPoseableMessageEvent(
						character = Character(id = data.get("char", {}).get("id"),
						                      bot = self),
						message = data.get("msg"),
						pose = data.get("pose"),
						puppeteer = Character(id = data.get("puppeteer", {}).get("id"),
						                      bot = self))
			
			elif event_class == TargetedCharacterMessageEvent:
				event.event = TargetedCharacterMessageEvent(
						character = Character(id = data.get("char", {}).get("id"),
						                      bot = self),
						
						message = data.get("msg"),
						target = Character(id = data.get("target", {}).get("id"),
						                   bot = self),
						ooc = data.get("ooc"),
						pose = data.get("pose"),
						puppeteer = Character(id = data.get("puppeteer", {}).get("id"),
						                      bot = self),
						targets = [Character(id = target.get("id"),
						                     bot = self) for target in data.get("targets", [])])
			
			elif event_class == TargetRoomMessageEvent:
				event.event = TargetRoomMessageEvent(
						character = Character(id = data.get("char", {}).get("id"),
						                      bot = self),
						message = data.get("msg"),
						target_room = Room(data.get("targetRoom", {}).get("name"),
						                   data.get("targetRoom", {}).get("id")),
						puppeteer = Character(id = data.get("puppeteer", {}).get("id"),
						                      bot = self))
			
			elif event_class == TargetedCharacterEvents:
				event.event = TargetedCharacterEvents(
						character = Character(id = data.get("char", {}).get("id"),
						                      bot = self),
						target = Character(id = data.get("target", {}).get("id"),
						                   bot = self),
						puppeteer = Character(id = data.get("puppeteer", {}).get("id"),
						                      bot = self))
			
			else:
				self.log_error(f"Error - Unknown Event Class - {event_class}")
				return
			
			# check if the character is the bot
			if event.event.character.id == self.id:
				return
			
			# check if the character.id is in the self.waiting_for_response_from dictionary
			if event.event.character.id in self.waiting_for_response_from:
				# check the type is not a TargetedCharacterEvents
				if event_class != TargetedCharacterEvents:
					self.waiting_for_response_from[event.event.character.id] = event.event.message
					# we want to return here as we don't want to call the on_calls
					return
			
			for func in self.on_calls[event.type]:
				func(event)
		
		@self.client.on("open")
		def on_open():
			self.log_info("Connection opened")
			
			# protocol version
			self.version = self.get_version()
			
			# authenticate
			self.authenticate()
			
			# get the bot
			self.get_bot()
			
			# subscribe to all required resources
			self.subscribe_to_all()
			
			# control the bot
			self.control_bot()
			
			# wake up the bot
			self.wake_up()
			
			self.booted = True
		
		# connect to the server
		if not self.client.running:
			self.client.connect()
		
		while not self.booted:
			time.sleep(0.1)
	
	def wake_up(self) -> None:
		self.log_info("Waking up bot")
		wake_up = self.client.call("core", "char", self.id, "ctrl", "wakeup").wait()
		self.get_error(wake_up, False)
		self.log_success("Successfully woke up bot")
	
	def say(self, message: any) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.say","params":{"msg":"message"}}
		self.log_info(f"Bot says - \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "say",
		                        msg = str(message)).send()
	
	def pose(self, message: any) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.pose","params":{"msg":"message"}}
		self.log_info(f"Bot poses - \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "pose",
		                        msg = str(message)).send()
	
	def ooc(self, message: any) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.ooc","params":{"msg":"message"}}
		self.log_info(f"Bot oocs - \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "ooc",
		                        msg = str(message)).send()
	
	def describe(self, message: any) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.describe","params":{"msg":"message"}}
		self.log_info(f"Bot describes - \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "describe",
		                        msg = str(message)).send()
	
	def message(self, target: Character, message: any, pose: bool = False, ooc: bool = False) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.message",
		# "params":{"charId":"{target_id}","msg":"hiya","pose":false, "ooc":false}}
		self.log_info(f"Bot messages - {target.full_name} with \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "message",
		                        charId = target.id,
		                        msg = str(message),
		                        pose = pose,
		                        ooc = ooc).send()
	
	def address(self, target: Character, message: any, pose: bool = False, ooc: bool = False) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.address",
		# "params":{"charId":"{target_id}","msg":"hiya","pose":false,"ooc":false}}
		self.log_info(f"Bot addresses - {target.full_name} with \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "address",
		                        charId = target.id,
		                        msg = str(message),
		                        pose = pose,
		                        ooc = ooc).send()
	
	def whisper(self, target: Character, message: any, pose: bool = False, ooc: bool = False) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.whisper",
		# "params":{"charId":"{target_id}","msg":"hiya","pose":false,"ooc":false}}
		self.log_info(f"Bot whispers - {target.full_name} with \"{str(message)}\"")
		return self.client.call("core", "char", self.id, "ctrl", "whisper",
		                        charId = target.id,
		                        msg = str(message),
		                        pose = pose,
		                        ooc = ooc).send()
	
	def mail(self, target: Character, message: any, pose: bool = False, ooc: bool = False) -> None:
		self.log_error("Bot mailing is not implemented!")
		self.log_info("Bots cannot mail for the time being due to requiring player ID's, not character ID's")
		self.log_info("This can be changed by allowing you to add your own player ID to the bot")
		self.log_info("However this is unlikely to be implemented due to the security risks")
		return None
	
	def summon(self, target: Character) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.summon",
		# "params":{"charId": "{target_id}"}}
		self.log_info(f"Bot summons - {target.full_name}")
		return self.client.call("core", "char", self.id, "ctrl", "summon",
		                        charId = target.id).send()
	
	def join(self, target: Character) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.join",
		# "params":{"charId":"{target_id}"}}
		self.log_info(f"Bot joins - {target.full_name}")
		return self.client.call("core", "char", self.id, "ctrl", "join",
		                        charId = target.id).send()
	
	def lead(self, target: Character):
		# {"id":0,"method":"call.core.char.{id}.ctrl.lead",
		# "params":{"charId":"{target_id}"}}
		self.log_info(f"Bot leads - {target.full_name}")
		return self.client.call("core", "char", self.id, "ctrl", "lead",
		                        charId = target.id).send()

	def follow(self, target: Character):
		# {"id":0,"method":"call.core.char.{id}.ctrl.follow",
		# "params":{"charId":"{target_id}"}}
		self.log_info(f"Bot follows - {target.full_name}")
		return self.client.call("core", "char", self.id, "ctrl", "follow",
		                        charId = target.id).send()
	
	def stop_lead(self, target: Character):
		# {"id":0,"method":"call.core.char.{id}.ctrl.stopLead",
		# "params":{"charId":"{target_id}"}}
		self.log_info(f"Bot stops leading - {target.full_name}")
		return self.client.call("core", "char", self.id, "ctrl", "stopLead",
		                        charId = target.id).send()

	def stop_follow(self):
		# {"id":0,"method":"call.core.char.{id}.ctrl.stopFollow",
		# "params":{}}
		self.log_info(f"Bot stops following")
		return self.client.call("core", "char", self.id, "ctrl", "stopFollow").send()
	
	def sleep(self) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.release","params":{}}
		self.log_info("Bot sleeps")
		return self.client.call("core", "char", self.id, "ctrl", "release").send()
	
	def ping(self) -> Request:
		# {"id":0,"method":"call.core.char.{id}.ctrl.ping","params":{}}
		self.log_info("Bot pings")
		return self.client.call("core", "char", self.id, "ctrl", "ping").send()
	
	def wait_for(self, character: Character, timeout: int = 10) -> str:
		# check if we're already waiting for a response from this character
		if character.id in self.waiting_for_response_from:
			self.log_error(f"Error: Bot Already Awaiting Response - {character.full_name}")
			return ""
		# add the character to the dict of characters we're waiting for a response from with value (the time)
		self.waiting_for_response_from[character.id] = time.time()
		# wait for a response from the character
		while type(self.waiting_for_response_from[character.id]) != str:
			# if the timeout is reached, return an empty string
			try:
				if time.time() - self.waiting_for_response_from[character.id] > timeout:
					self.log_info(f"Bot timed out waiting for a response from {character.full_name}")
					del self.waiting_for_response_from[character.id]
					return ""
			except TypeError:
				# another thread has changed the value of the dict entry to a string
				# this means we have a response from the character
				pass
		
		# get the response from the character
		response = self.waiting_for_response_from[character.id]
		# remove the character from the dict of characters we're waiting for a response from
		del self.waiting_for_response_from[character.id]
		# return the response
		return response
	
	def look_up_characters(self, character_name: str) -> list[Character]:
		# {"id":0,"method":"call.core.char.{id}.ctrl.lookUp",
		# "params":{"name":"{character_name}"}}
		names = character_name.split(" ")
		self.log_info(f"Bot looking up characters with the name {names[0]}")
		characters = self.client.call(self.rid, "lookupChars",
		                              name = names[0], extended = True).wait()
		
		self.log_success(characters)
		print("char: ", self.client.cache.get(f"core.char.ccicqnu9gbrk70s5rrdg"))
		
		# {'result': {'payload': {'chars': [{'id': 'ccqvh7m9gbrk70si43e0', 'name': 'Omi', 'surname': 'Lillabi',
		# 'species': 'Mouse', 'gender': 'Male', 'awake': True, 'lastAwake': 1690799543339}]}}, 'id': 8}
		
		if not self.get_error(characters, False):
			characters = [Character(id = character.get("id"),
			                        bot = self)
			              for character in characters.value()["result"]["payload"]["chars"]]
			
			if len(names) > 1:
				# filter out characters that don't have the second name
				self.log_info(f"Bot found {len(characters)} characters with the name {character_name}")
				return [character for character in characters if
				        character.surname.lower() == " ".join(names[1:]).lower()]
			
			else:
				self.log_info(f"Bot found {len(characters)} characters with the name {character_name}")
				return characters
		
		else:
			self.log_error(f"Error - Bot Failed To Look Up Characters - {character_name}")
			return []
	
	def use_exit(self, exit: Exit):
		# {"id":0,"method":"call.core.char.{id}.ctrl.useExit","params":{"exitId":"{exit_id}"}}
		self.log_info(f"Bot uses exit - {exit.name}")
		self.client.call("core", "char", self.id, "ctrl", "useExit", exitId = exit.id).send()
	
	def get_room(self) -> Room:
		return Room(self, self.client.cache.get(f"core.char.{self.id}.owned", "inRoom", "id"))
	
	def kill(self) -> None:
		self.log_info("Killing bot")
		self.sleep()
		self.client.close()
		self.log_success("Successfully killed bot")
		sys.exit()


from src.Mucklet.ResClient import Request

try:
	from src.Mucklet.bot import Bot
except ImportError:
	pass


# Properties used by all events
class Character:
	def __init__(self, bot: "Bot", id: str) -> None:
		self.bot: "Bot" = bot
		self.id = id
	
	def __str__(self):
		return f"{self.full_name} ({self.id})"
	
	def __repr__(self):
		return str(self)
	
	def message(self, message: str, pose: bool = False, ooc: bool = False) -> Request:
		return self.bot.message(self, message, pose, ooc)
	
	def whisper(self, message: str, pose: bool = False, ooc: bool = False) -> Request:
		return self.bot.whisper(self, message, pose, ooc)
	
	def address(self, message: str, pose: bool = False, ooc: bool = False) -> Request:
		return self.bot.address(self, message, pose, ooc)
	
	def summon(self):
		return self.bot.summon(self)
	
	def join(self):
		return self.bot.join(self)
	
	def get_avatar(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "avatar"))
	
	def get_awake(self):
		return bool(self.bot.client.cache.get(f"core.char.{self.id}", "awake"))
	
	def get_gender(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "gender"))
	
	def get_idle(self):
		return int(self.bot.client.cache.get(f"core.char.{self.id}", "idle"))
	
	def get_last_awake(self):
		return int(self.bot.client.cache.get(f"core.char.{self.id}", "lastAwake"))
	
	def get_name(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "name"))
	
	def get_species(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "species"))
	
	def get_state(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "state"))
	
	def get_status(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "status"))
	
	def get_surname(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "surname"))
	
	def get_tags(self):
		return list(self.bot.client.cache.get(f"core.char.{self.id}", "tags"))
	
	def get_type(self):
		return str(self.bot.client.cache.get(f"core.char.{self.id}", "type"))
	
	def get_full_name(self):
		return str(self.name + " " + self.surname)
	
	name = property(get_name)
	surname = property(get_surname)
	avatar = property(get_avatar)
	awake = property(get_awake)
	gender = property(get_gender)
	idle = property(get_idle)
	last_awake = property(get_last_awake)
	species = property(get_species)
	state = property(get_state)
	status = property(get_status)
	tags = property(get_tags)
	type = property(get_type)
	
	full_name = property(get_full_name)


class Area:
	def __init__(self, bot: "Bot", id: str):
		self.bot = bot
		self.id = id
	
	def __str__(self):
		return f"{self.name} {self.short_description} ({self.id})"
	
	def __repr__(self):
		return str(self)
	
	def get_about(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "about")
	
	def get_children(self):
		return [Room(self.bot, room) for room in self.bot.client.cache.get(f"core.area.{self.id}.details", "children")]
	
	def get_image(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "image")
	
	def get_map_x(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "mapX")
	
	def get_map_y(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "mapY")
	
	def get_owner(self):
		return Character(self.bot, self.bot.client.cache.get(f"core.area.{self.id}.details", "owner", "id"))
	
	def get_parent(self):
		return Area(self.bot, self.bot.client.cache.get(f"core.area.{self.id}.details", "parent", "id"))
	
	def get_pop(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "pop")
	
	def get_private(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "private")
	
	def get_prv(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "prv")
	
	def get_rules(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "rules")
	
	def get_short_description(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "shortDesc")
	
	def get_name(self):
		return self.bot.client.cache.get(f"core.area.{self.id}.details", "name")
	
	about = property(get_about)
	children = property(get_children)
	image = property(get_image)
	map_x = property(get_map_x)
	map_y = property(get_map_y)
	name = property(get_name)
	owner = property(get_owner)
	parent = property(get_parent)
	pop = property(get_pop)
	private = property(get_private)
	prv = property(get_prv)
	rules = property(get_rules)
	short_description = property(get_short_description)


class Room:
	def __init__(self, bot: "Bot", id: str) -> None:
		self.bot = bot
		self.id = id
	
	def __str__(self):
		return f"{self.name} ({self.id})"
	
	def __repr__(self):
		return str(self)
	
	def get_area(self):
		return Area(self.bot, self.bot.client.cache.get(f"core.room.{self.id}.details", "area", "id"))
	
	def get_autosweep(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "autosweep")
	
	def get_autosweep_delay(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "autosweepDelay")
	
	def get_characters(self):
		return [Character(self.bot, self.bot.client.cache.get(character.get("rid"), "id"))
		        for character in self.bot.client.cache.get(f"core.room.{self.id}.details", "chars")]
	
	def get_description(self) -> str:
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "desc")
	
	def get_exits(self) -> list["Exit"]:
		return [Exit(self.bot, self.bot.client.cache.get(exit.get("rid"), "id"))
		        for exit in self.bot.client.cache.get(f"core.room.{self.id}.details", "exits")]
	
	def get_image(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "image")
	
	def get_is_dark(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "isDark")
	
	def get_is_home(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "isHome")
	
	def get_is_quiet(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "isQuiet")
	
	def get_is_teleport(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "isTeleport")
	
	def get_map_x(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "mapX")
	
	def get_map_y(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "mapY")
	
	def get_name(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "name") or \
			self.bot.client.cache.get(f"core.room.{self.id}", "name")
	
	def get_owner(self):
		return Character(self.bot, self.bot.client.cache.get(f"core.room.{self.id}.details", "owner", "id"))
	
	def get_pop(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "pop")
	
	def get_private(self):
		return self.bot.client.cache.get(f"core.room.{self.id}.details", "private")
	
	area = property(get_area)
	autosweep = property(get_autosweep)
	autosweep_delay = property(get_autosweep)
	characters = property(get_characters)
	description = property(get_description)
	exits = property(get_exits)
	image = property(get_image)
	is_dark = property(get_is_dark)
	is_home = property(get_is_home)
	is_quiet = property(get_is_quiet)
	is_teleport = property(get_is_teleport)
	map_x = property(get_map_x)
	map_y = property(get_map_y)
	name = property(get_name)
	owner = property(get_owner)
	pop = property(get_pop)
	private = property(get_private)


class Exit:
	def __init__(self, bot: "Bot", id: str):
		self.bot = bot
		self.id = id
	
	def __str__(self):
		return f"{self.name} {self.keys} ({self.id})"
	
	def __repr__(self):
		return str(self)
	
	def get_keys(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "keys", "data") or \
			self.bot.client.cache.get(f"core.exit.{self.id}", "keys", "data")
	
	def get_name(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "name") or \
			self.bot.client.cache.get(f"core.exit.{self.id}", "name")
	
	def get_arrive_message(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "arriveMsg")
	
	def get_leave_message(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "leaveMsg")
	
	def get_created(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "created")
	
	def get_hidden(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "hidden")
	
	def get_target_room(self):
		return Room(self.bot, self.bot.client.cache.get(f"core.exit.{self.id}.details", "targetRoom", "id"))
	
	def get_travel_message(self):
		return self.bot.client.cache.get(f"core.exit.{self.id}.details", "travelMsg")
	
	keys = property(get_keys)
	name = property(get_name)
	arrive_message = property(get_arrive_message)
	leave_message = property(get_leave_message)
	created = property(get_created)
	hidden = property(get_hidden)
	target_room = property(get_target_room)
	travel_message = property(get_travel_message)


class EventBase:
	def __init__(self, id: int, type: str, time: float, sig: str):
		self.id = id
		self.type = type
		self.time = time
		self.sig = sig
		self.event = None
	
	def __str__(self):
		return f"{self.type} event {self.id} at {self.time} with signature {self.sig}:\n\tEvent: {self.event}"


# say, pose, wakeup, sleep, leave, arrive, describe, action
class CharacterMessageEvent:
	def __init__(self, character: Character, message: str, puppeteer: Character | None = None):
		self.character = character
		self.message = message
		self.puppeteer = puppeteer
	
	def __str__(self):
		return f"{self.character.name} {self.character.surname} says \"{self.message}\""


# ooc
class CharacterPoseableMessageEvent:
	def __init__(self, character: Character, message: str, pose: bool = False, puppeteer: Character | None = None):
		self.character = character
		self.message = message
		self.pose = pose
		self.puppeteer = puppeteer
	
	def __str__(self):
		return f"{self.character.name} {self.character.surname} says \"{self.message}\""


# whisper, message, warn, mail, address, controlRequest
class TargetedCharacterMessageEvent:
	def __init__(self, character: Character,
	             message: str,
	             target: Character,
	             ooc: bool = False,
	             pose: bool = False,
	             puppeteer: Character | None = None,
	             targets: list[Character] | None = None):
		self.character = character
		self.message = message
		self.target = target
		self.ooc = ooc
		self.pose = pose
		self.puppeteer = puppeteer
		self.targets = targets
		if self.targets is None:
			self.targets = []
		self.targets.append(self.target)
	
	def __str__(self):
		return f"{self.character.name} {self.character.surname} says \"{self.message}\" to " \
		       f"{self.target.name} {self.target.surname}"


# travel
class TargetRoomMessageEvent:
	def __init__(self, character: Character, message: str, target_room: Room, puppeteer: Character | None = None):
		self.character = character
		self.message = message
		self.target = target_room
		self.puppeteer = puppeteer
	
	def __str__(self):
		return f"{self.character.name} {self.character.surname} says \"{self.message}\" to {self.target.name}"


# summon, join, leadRequest, followRequest, follow, stopFollow, stopLead
class TargetedCharacterEvents:
	def __init__(self, character: Character, target: Character, puppeteer: Character | None = None):
		self.character = character
		self.target = target
		self.puppeteer = puppeteer
	
	def __str__(self):
		return f"{self.character.name} {self.character.surname} is targeting {self.target.name} {self.target.surname}"



import websocket
import json
import time
import threading
from src.Mucklet.log import Logger


class Request:
	iteration = 0
	time_out_period = 10
	
	def __init__(self, _connection, *method, notification: bool = False, **params):
		self.connection = _connection
		Request.iteration += 1
		self.id = Request.iteration
		self.method = ".".join(method)
		self.params = params or {}
		self.pending_data = {"jsonrpc": "2.0",
		                     "method" : self.method,
		                     "params" : self.params, }
		
		self.notification = notification
		
		# if the request is a notification (i.e. it does not require a response)
		if not self.notification:
			self.pending_data.update({"id": self.id})
		
		self.received_data = None
		self.sent_data = None
		self.sent_time = None
		self.recieved_time = None
		self.timed_out = False
	
	def send(self) -> "Request":
		if self.sent_data:
			self.connection.log_error("Message has already been sent")
			return self
		if self.received_data:
			self.connection.log_error("Message has already been received")
			return self
		if not self.connection.running:
			self.connection.log_error("Connection is not running")
			return self
		
		if not self.notification:
			# add the message to the promises
			self.connection.promises[self.id] = self
		# send the message
		self.connection.ws.send(json.dumps(self.pending_data))
		self.sent_data = self.pending_data.copy()
		self.sent_time = time.time()
		Logger.log_if_not_shown(f"Delivered - {self.sent_data}", (100, 200, 100))
		return self
	
	def resend(self) -> None:
		self.sent_data = None
		self.received_data = None
		self.send()
	
	def receive(self, data: dict) -> None:
		self.received_data = data.copy()
		self.pending_data = None
		# remove the message from the promises
		self.kill()
		self.recieved_time = time.time()
		
	def wait(self, time_out_period: int = None) -> "Request":
		if self.sent_data is None:
			# send the message
			self.send()
		time_out_period = time_out_period or Request.time_out_period
		# while the message has not timed out
		while time.time() - self.sent_time < time_out_period:
			# if the message has been received
			if self.received_data:
				# return self
				return self
			# sleep for 0.1 seconds
			time.sleep(0.1)
		# if the message has timed out
		self.connection.log_error(f"Message {self.id} has timed out", False)
		self.timed_out = True
		return self
	
	def __str__(self) -> str:
		ping = (self.recieved_time - self.sent_time) * 1000 if self.recieved_time else None
		ping = f"{ping:.2f}" if ping else "N/A"
		if self.received_data:
			return f"Message {self.id} [{ping}ms]:\n   --> {self.sent_data}\n   <-- {self.received_data}"
		if self.sent_data:
			return f"Message {self.id}: sent - {self.sent_data}, not received"
		return f"Message {self.id}: not sent, not received"
	
	def kill(self) -> None:
		try:
			self.connection.promises.pop(self.id)
		except KeyError:
			self.connection.log_error(f"Message {self.id} has already been killed", False)
	
	def value(self) -> dict:
		if not self.received_data:
			self.connection.log_error(f"Message {self.id} has not been received, now waiting for the data", False)
			self.wait()
		return self.received_data


class Cache:
	def __init__(self, client: "ResClient", data: any = None):
		self.client = client
		if data:
			self.data = data
		else:
			self.data = {}
		
		self.errored = set()
	
	def get(self, *rids: str, default: any = None) -> any:
		# follow the keys to the end, if there is a rid then start from the head with the rid and continue
		
		position = self.data
		for rid in rids:
			try:
				position = position[rid]
				if type(position) is dict and "rid" in position:
					position = self.get(position["rid"])
			except KeyError:
				if rid in self.errored:
					return default
				# we can try and subscribe to the resource
				resource = self.client.subscribe(rid).wait()
				if "error" in resource.value():
					self.errored.add(rid)
					return default
				try:
					position = self.get(rid)
				except KeyError:
					self.client.log_error(rid)
					self.client.log_error(str(resource))
					return default
		
		return position
	
	def set(self, rid: str, value: any) -> None:
		self.data[rid] = value
	
	def change(self, rid: str, values: dict) -> None:
		for key, value in values.items():
			if value == {"action": "delete"}:
				self.data[rid].pop(key)
			else:
				self.data[rid][key] = value
	
	def add(self, rid: str, index: int, value: dict) -> None:
		self.data.get(rid).insert(index, value)
	
	def remove(self, rid: str, index: int) -> None:
		self.data.get(rid).pop(index)
	
	def __getitem__(self, item: any) -> any:
		return self.get(item)
	
	def __setitem__(self, key: str, value: any) -> None:
		self.set(key, value)
	
	def __delitem__(self, key: str) -> None:
		self.data.pop(key)
	
	def __str__(self) -> str:
		return str(self.data)
	
	def __repr__(self) -> str:
		return str(self.data)

	def __contains__(self, item: str) -> bool:
		return item in self.data


class Connection:
	def __init__(self, host: str, origin: str):
		self.host = host
		self.origin = origin
		self.ws = websocket.WebSocketApp(self.host,
		                                 on_message = self.on_message,
		                                 on_error = self.on_error,
		                                 on_close = self.on_close,
		                                 on_open = self.on_open, )
		self.running = False
		self.promises = {}
		self.log_info = Logger(format_color = (80, 80, 210), bold = False)
		self.log_error = Logger(format_color = (210, 80, 80), bold = False)
		self.log_success = Logger(format_color = (80, 210, 80), bold = False)
		
		self.on_events = {
			"message": lambda message: self.log_info(f"Received - {message}", False),
			"error"  : lambda error: self.log_error(f"Error - {error}", False),
			"close"  : lambda: self.log_info("Connection closed", False),
			"open"   : lambda: self.log_info("Connection opened", False),
		}
		
	def connect(self) -> "Connection":
		def run_forever():
			self.ws.run_forever(origin = self.origin)
		
		threading.Thread(target = run_forever).start()
		while not self.running:
			time.sleep(0.1)
		return self
	
	def on_message(self, ws, message):
		message = json.loads(message)
		if "id" in message:
			if message["id"] in self.promises:
				self.promises[message["id"]].receive(message)
			else:
				self.log_error(f"Recieved message with id {message['id']} that was not sent by this client", False)
		threading.Thread(target = self.on_events["message"], args = (message,)).start()
		
	def on_error(self, ws, error):
		threading.Thread(target = self.on_events["error"], args = (error,)).start()
	
	def on_close(self, ws):
		self.running = False
		threading.Thread(target = self.on_events["close"]).start()
	
	def on_open(self, ws):
		self.running = True
		threading.Thread(target = self.on_events["open"]).start()
	
	def request(self, *args, notification = False, **kwargs) -> Request:
		return Request(self, *args, **kwargs, notification = notification)
	
	def emulate(self, **message) -> None:
		if "id" in message:
			if message["id"] in self.promises:
				self.promises[message["id"]].receive(message)
			else:
				self.log_error(f"Recieved message with id {message['id']} that was not sent by this client", False)
		threading.Thread(target = self.on_events["message"], args = (message,)).start()


class ResClient(Connection):
	def __init__(self, host: str, origin: str):
		super().__init__(host, origin)
		
		self.cache = Cache(self)
		self.cache.head = self.cache

	def on_message(self, ws, response):
		response = json.loads(response)
		if "id" in response:
			if response["id"] in self.promises:
				self.promises[response["id"]].receive(response)
			else:
				self.log_error(f"Recieved message with id {response['id']} that was not sent by this client", False)
		self.process_response(response)

	def process_response(self, response: dict) -> None:
		Logger.log_if_not_shown(str(response))
		
		if "result" in response:
			result = response.get("result", {})
			if "models" in result:
				self.cache_dict(result.get("models", {}))
				
			if "collections" in result:
				self.cache_dict(result.get("collections", {}))
			
			if "errors" in result:
				self.display_errors(result.get("errors", {}))
		
		if "data" in response:
			data = response.get("data", {})
			if "models" in data:
				self.cache_dict(data.get("models", {}))
			
			if "collections" in data:
				self.cache_dict(data.get("collections", {}))
			
			if "errors" in data:
				self.display_errors(data.get("errors", {}))
		
		if "event" in response:
			if "data" not in response:
				self.log_error(f"Event {response['event']} did not contain any data", False)
				return
			data = response.get("data", {})
			event = response.get("event", {}).split(".")
			rid = ".".join(event[:-1])
			event = event[-1]
			
			if event == "change":
				self.cache.change(rid, data.get("values", {}))
			
			if event == "add":
				self.cache.add(rid, data.get("idx", 0), data.get("value", {}))
			
			if event == "remove":
				self.cache.remove(rid, data.get("idx", 0))
			
			if event in ["add", "change", "create", "delete", "patch", "reset", "reaccess", "remove", "unsubscribe"]:
				return
			
			else:
				threading.Thread(target = self.on_events["message"], args = (response,)).start()
	
	def display_errors(self, errors: dict) -> None:
		for service, error in errors.items():
			self.log_error(f"{service} - {error.get('code')} - {error.get('message')}", False)
	
	def cache_dict(self, values: dict) -> None:
		for rid, value in values.items():
			self.cache.set(rid, value)
	
	def version(self, protocol: str) -> Request:
		"""
		Version request
			method: version
			
			Version requests are sent by the client to tell which RES protocol
			version it supports, and to get information on what protocol version
			the gateway supports.
			
			The request SHOULD be the first request sent by the client after an
			established connection.
			
			If not sent, or if the protocol property is omitted in the request,
			the gateway SHOULD assume version v1.1.x.
			
			Parameters
				The request parameters are optional.
				If not omitted, the parameters object SHOULD have the following
				property:
			
				protocol
					The RES protocol version supported by the client.
					MUST be a string in the format "[MAJOR].[MINOR].[PATCH]". Eg. "1.2.3".
			
			Result
				protocol
					The RES protocol version supported by the gateway.
					MUST be a string in the format "[MAJOR].[MINOR].[PATCH]". Eg. "1.2.3".
			
			Error
				A system.unsupportedProtocol error response will be sent if the gateway
				cannot support the client protocol version.
				A system.invalidRequest error response will be sent if the gateway only
				supports RES Protocol v1.1.1 or below, prior to the introduction of the
				version request.
		"""
		return self.request("version", protocol = protocol)
	
	def subscribe(self, *rid):
		"""
		Subscribe request
			method: subscribe.<resourceID>
			
			Subscribe requests are sent by the client to subscribe to a resource.
			The request has no parameters.
			
			Result
				models
					Resource set models.
					May be omitted if no new models were subscribed.
				
				collections
					Resource set collections.
					May be omitted if no new collections were subscribed.
				
				errors
					Resource set errors.
					May be omitted if no subscribed resources encountered errors.
			
			Error
				An error response will be sent if the resource couldn't be subscribed to.
				Any resource reference that fails will not lead to an error response, but
				the error will be added to the resource set errors.
		"""
		return self.request("subscribe", *rid)
	
	def unsubscribe(self, *rid, count = 1):
		"""
		Unsubscribe request
			method: unsubscribe.<resourceID>
			
			Unsubscribe requests are sent by the client to unsubscribe to previous direct
			subscriptions.
			
			The resource will only be considered unsubscribed when there are no more direct
			or indirect subscriptions.
			
			If the count property is omitted in the request, the value of 1 is assumed.
			
			
			Parameters
				The request parameters are optional.
				If not omitted, the parameters object SHOULD have the following property:
			
				count
					The number of direct subscriptions to unsubscribe to.
					MUST be a number greater than 0.
				
			Result
				The result has no payload.
			
			Error
				An error response with code system.noSubscription will be sent if the
				resource has no direct subscription, or if count exceeds the number of
				direct subscriptions. If so, the number of direct subscriptions will be
				unaffected.
		"""
	
	def get(self, *rid):
		"""
		Get requests are sent by the client to get a resource without making a subscription.
		
		method: get.<resourceID>
		
		Parameters
			The request has no parameters.
		
		Result
			models
				Resource set models.
				May be omitted if no new models were retrieved.
			
			collections
				Resource set collections.
				May be omitted if no new collections were retrieved.
			
			errors
				Resource set errors.
				May be omitted if no retrieved resources encountered errors.
		
		Error
			An error response will be sent if the resource couldn't be retrieved.
			Any resource reference that fails will not lead to an error response, but the
			error will be added to the resource set errors.
		"""
		return self.request("get", *rid)
	
	def call(self, *rid: str, **kwargs):
		"""
		Call request
			method: call.<resourceID>.<resourceMethod>
			
			Call requests are sent by the client to invoke a method on the resource.
			The response may either contain a result payload or a resource ID.
			
			In case of a resource ID, the resource is considered directly subscribed.
			
			Parameters
				The request parameters are defined by the service.
			
			Result
				The result is an object with the following members:
				
				payload
					Result payload as defined by the service.
					MUST be omitted if rid is set.
					MUST NOT be omitted if rid is not set.
				
				rid
					Resource ID of subscribed resource.
					MUST be omitted if payload is set.
				
				models
					Resource set models.
					May be omitted if no new models were subscribed.
					MUST be omitted if payload is set.
				
				collections
					Resource set collections.
					May be omitted if no new collections were subscribed.
					MUST be omitted if payload is set.
				
				errors
					Resource set errors.
					May be omitted if no subscribed resources encountered errors.
					MUST be omitted if payload is set.
			
			Error
				An error response will be sent if the method couldn't be called,
				or if the method was called, but an error was encountered.
		"""
		return self.request("call", *rid, **kwargs)
	
	def auth(self, *rid, **params):
		"""
		Auth request
		method: auth.<resourceID>.<resourceMethod>
		
		Auth requests are sent by the client to authenticate the client connection.
		The response may either contain a result payload or a resource ID.
		
		In case of a resource ID, the resource is considered directly subscribed.
		
		Parameters
			The request parameters are defined by the service.
		
		Result
			The result is an object with the following members:
			
			payload
				Result payload as defined by the service.
				MUST be omitted if rid is set.
				MUST NOT be omitted if rid is not set.
			
			rid
				Resource ID of subscribed resource.
				MUST be omitted if payload is set.
			
			models
				Resource set models.
				May be omitted if no new models were subscribed.
				MUST be omitted if payload is set.
			
			collections
				Resource set collections.
				May be omitted if no new collections were subscribed.
				MUST be omitted if payload is set.
		Error
			An error response will be sent if the resource could not be created,
			or if an error was encountered retrieving the newly created resource.
		"""
		return self.request("auth", *rid, **params)
	
	def on(self, event: str):
		if event not in self.on_events:
			self.log_error(f"Event '{event}' is not supported. Supported events are: {', '.join(self.on_events)}")
			return
		
		def decorator(func):
			self.on_events[event] = func
			return func
		
		return decorator
	
	def close(self):
		self.ws.close()
		self.promises.clear()
	
	def new(self):
		cls = self.__class__
		result = cls(self.host, self.origin)
		return result



import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "Mucklet",
    version = "1.0.4",
    author = "Ico Twilight",
    author_email = "Ico@duck.com",
    description = "A package for creating and using bots for the Mucklet role-playing service.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/IcoTwilight/PythonMuckletBot",
    project_urls = {
        "Bug Tracker": "https://github.com/IcoTwilight/PythonMuckletBot/issues",
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    install_requires=[
          'websocket-client',
    ],
    python_requires = ">=3.6"
)



class Logger:
	iteration = 0
	shown = True
	
	def __init__(self, format_color: tuple[int, int, int] = (255, 50, 50), bold: bool = True):
		self.format_color = format_color
		self.bold = bold
		# turn the format color into an ansi escape code
		self.format_code = "\033[38;2;{};{};{}m".format(*format_color)
		# add bold if needed
		if bold: self.format_code += "\033[1m"
		# add the escape code
		self.escape_code = "\033[0m"
	
	def __call__(self, _message: any, _raise: bool = False) -> None:
		if not Logger.shown: return
		# print the message to the console
		Logger.iteration += 1
		if _raise: raise Exception(_message)
		print(f"{self.format_code}{Logger.iteration}: {str(_message)}{self.escape_code}")
	
	@classmethod
	def reset(cls) -> None:
		cls.iteration = 0
	
	@classmethod
	def hide(cls) -> None:
		cls.shown = False
	
	@classmethod
	def show(cls) -> None:
		cls.shown = True
	
	@classmethod
	def toggle(cls) -> None:
		cls.shown = not cls.shown
	
	@classmethod
	def log_if_not_shown(cls, _message: str, color = (255, 255, 255)) -> None:
		color_prefix = "\033[38;2;{};{};{}m".format(*color)
		color_suffix = "\033[0m"
		if not cls.shown:
			print(f"{color_prefix}{_message}{color_suffix}")
